def apply(hub, raw, raw_cli, cli, cli_args, os_vars, configs):
    # Defaults (raw)
    # Config files (configs)
    # OS (os_vars)
    # CLI (cli_args)
    ret = {}
    cli_map = {}
    for imp in raw:
        ret[imp] = {}
        cli_map[imp] = {}
        for key, data in raw[imp]["CONFIG"].items():
            if "default" in data:
                ret[imp][key] = data["default"]
    # TODO: This assumes that we are using the namespace approach,
    # This makes the config structure the easiest, meaning that components are
    # namespaced by the user.
    # Some other, additioanl pattern could be added and this chunk could
    # be made pluggable.
    for imp in configs:
        if imp not in ret:
            ret[imp] = {}
        for key in configs[imp]:
            ret[imp][key] = configs[imp][key]
    for imp in os_vars:
        for key in os_vars[imp]:
            ret[imp][key] = os_vars[imp][key]
    for key in cli_args:
        if key in raw_cli:
            if "source" in raw_cli[key]:
                ret[raw_cli[key]["source"]][key] = cli_args[key]
            else:
                ret[cli][key] = cli_args[key]
    ret = hub.config.roots.apply(ret, raw, cli)
    return ret



# The render process for the cli does NOT use the rend project for a couple
# of reasons
# 1. Config needs to be very early in the startup, therefore it cannot use
#    asyncio, all rend funcs use asyncio
# 2. Config will nto allow for template wrapping as the render is just a
#    single command line render


def process(hub, renderer, value):
    """
    Take a renderer adn a value, process it, and return the processed value
    """
    print(renderer)
    print(value)
    return getattr(hub, f"config.render.{renderer}.render")(value)



# -*- coding: utf-8 -*-
"""
Define the yaml loader interface
"""

# Import third party libs
try:
    import toml

    HAS_TOML = True
except ImportError:
    HAS_TOML = False

__virtualname__ = "toml"
# __contracts__ = [__virtualname__]


def __virtual__(hub):
    if HAS_TOML:
        return True
    return (False, "TOML could not be loaded")


def load(hub, path):
    """
    use toml to read in a file
    """
    try:
        with open(path, "rb") as fp_:
            return toml.load(fp_.read())
    except FileNotFoundError:
        pass
    return {}


def render(hub, val):
    """
    Take the string and render it in json
    """
    return toml.loads(val)



"""
The os sub is used to gather configuration options from the OS facility
to send configuration options into applications.
"""
# Import python libs
import os


def gather(hub, raw):
    """
    Collect the keys that need to be found and pass them to the
    os specific loaded plugin
    """
    ret = {}
    for imp in raw:
        ret[imp] = {}
        for sec in hub.config.CONFIG_SECTIONS:
            if sec not in raw[imp]:
                continue
            for key in raw[imp][sec]:
                osvar = raw[imp][sec][key].get("os", None)
                if osvar is not None:
                    val = hub.config.os.system.collect(osvar)
                    if val is not None:
                        ret[imp][key] = val
    return ret



# Import python libs
import copy


def add(hub, cli, raw):
    """
    Inject any additional arguments onto the raw input that the config system
    wants to add, such as logging and version checking.
    """
    keys = set()
    if "CONFIG" not in raw[cli]:
        raw[cli]["CONFIG"] = {}
    # Merge in log configs
    dst = hub.config.log.init.conf(cli)
    keys.update(dst.keys())
    hub.pop.dicts.update(dst, raw[cli]["CONFIG"])
    raw[cli]["CONFIG"] = dst
    # Merge in Version configs
    dst = copy.deepcopy(hub.config.version.CONFIG)
    keys.update(dst.keys())
    hub.pop.dicts.update(dst, raw[cli]["CONFIG"])
    raw[cli]["CONFIG"] = dst
    # Apply the CLI_CONFIG
    for key in sorted(keys):
        if key not in raw[cli]["CLI_CONFIG"]:
            raw[cli]["CLI_CONFIG"][key] = {}
    return raw



def __init__(hub):
    """
    Load the subdirs for conf
    """
    hub.pop.sub.load_subdirs(hub.config, recurse=True)
    hub.config.ARGS = {}
    hub.config.SECTIONS = ("CONFIG", "CLI_CONFIG", "SUBCOMMANDS", "DYNE")
    hub.config.CONFIG_SECTIONS = ("CONFIG", "CLI_CONFIG")



# -*- coding: utf-8 -*-
"""
Translate an options data structure into command line args
"""

# Import python libs
import sys
import inspect
import argparse
from typing import Any, Dict, List, Tuple

DEFAULT = "247192870545555772379508883198090009006"


def _keys(opts):
    """
    Return the keys in the right order
    """
    return sorted(opts, key=lambda k: (opts[k].get("display_priority", sys.maxsize), k))


def gather(
    hub, raw: Dict[str, Any], cli: str, parse_cli: bool
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Return the cli arguments as they are parsed
    """
    if not parse_cli:
        return {}, {}
    raw_cli = hub.config.args.get_cli(raw, cli)
    hub.config.args.init_parser()
    hub.config.args.subparsers(raw, cli)
    defaults = hub.config.args.setup(raw_cli)
    cli_args = hub.config.args.parse()
    cli_args = hub.config.args.render(cli_args, raw_cli)
    cli_args = hub.config.args.clean_defaults(cli_args)
    return cli_args, raw_cli


def clean_defaults(hub, cli_args: Dict[str, Any]) -> Dict[str, Any]:
    """
    If anyone did not pass in an argument then the key will match the
    bad default and needs to be removed
    """
    ret = {}
    for key, val in cli_args.items():
        if str(val) != DEFAULT:
            ret[key] = val
    return ret


def init_parser(hub):
    if "parser" not in hub.config.ARGS:
        # Instantiate the parser
        hub.config.ARGS["parser"] = argparse.ArgumentParser()


def get_cli(hub, raw: Dict[str, Any], cli: str) -> Dict[str, Any]:
    """
    Gather the arguments that need to be parsed by the CLI
    """
    ret = {}
    main = raw.get(cli, {}).get("CLI_CONFIG")
    main_raw = raw.get(cli, {}).get("CONFIG")

    for key, data in main.items():
        ret[key] = {}
        hub.pop.dicts.update(ret[key], data)
        if key in main_raw:
            hub.pop.dicts.update(ret[key], main_raw[key])
        if "source" in data:
            src = raw.get(data["source"], {}).get("CONFIG", {}).get(key)
            if src is not None:
                hub.pop.dicts.update(ret[key], src)
        if "default" in ret[key]:
            ret[key]["default"] = DEFAULT
    ret.update(hub.config.version.CONFIG)
    return ret


def subparsers(hub, raw: Dict[str, Any], cli: str) -> bool:
    """
    Look over the data and extract and set up the subparsers for subcommands
    """
    subs = raw.get(cli, {}).get("SUBCOMMANDS")
    if not subs:
        return True
    hub.config.ARGS["sub"] = hub.config.ARGS["parser"].add_subparsers(
        dest="_subparser_"
    )
    hub.config.ARGS["subs"] = {}
    for arg in _keys(subs):
        if arg in ("_argparser_",):
            continue
        comps = subs[arg]
        kwargs = {}
        if "help" in comps:
            kwargs["help"] = comps["help"]
        if "desc" in comps:
            kwargs["description"] = comps["desc"]
        hub.config.ARGS["subs"][arg] = hub.config.ARGS["sub"].add_parser(arg, **kwargs)
    return True


def setup(hub, raw_cli: Dict[str, Any]) -> Dict[str, Any]:
    """
    Take in a pre-defined dict and translate it to args

    opts dict:
        <arg>:
            [group]: foo
            [default]: bar
            [action]: store_true
            [options]: # arg will be turned into --arg
              - '-A'
              - '--cheese'
            [choices]:
              - foo
              - bar
              - baz
            [nargs]: +
            [type]: int
            [dest]: cheese
            help: Some great help message
    """
    # TODO: This should be broken up
    defaults = {}
    groups = {}
    ex_groups = {}
    for arg in _keys(raw_cli):
        if arg in ("_argparser_",):
            continue
        comps = raw_cli[arg]
        positional = comps.pop("positional", False)
        if positional:
            args = [arg]
        else:
            long_opts = set(["--{}".format(arg.replace("_", "-"))])
            short_opts = set()
            for o_str in comps.get("options", []):
                if not o_str.startswith("--") and o_str.startswith("-"):
                    short_opts.add(o_str)
                elif o_str.startswith("--"):
                    long_opts.add(o_str)
                elif len(o_str) == 1:
                    short_opts.add(f"-{o_str}")
                else:
                    long_opts.add(f"--{o_str}")
            args = sorted(list(short_opts.union(long_opts)))
        kwargs = {}
        kwargs["action"] = action = comps.get("action", None)

        if action is None:
            # Non existing option defaults to a StoreAction in argparse
            action = hub.config.ARGS["parser"]._registry_get("action", action)

        if isinstance(action, str):
            signature = inspect.signature(
                hub.config.ARGS["parser"]._registry_get("action", action).__init__
            )
        else:
            signature = inspect.signature(action.__init__)

        for param in signature.parameters:
            if param == "self" or param not in comps:
                continue
            if param == "dest":
                kwargs["dest"] = comps.get("dest", arg)
                continue
            if param == "help":
                kwargs["help"] = comps.get("help", "THIS NEEDS SOME DOCUMENTATION!!")
                continue
            if param == "default":
                defaults[comps.get("dest", arg)] = comps[param]
            kwargs[param] = comps[param]

        if "group" in comps:
            group = comps["group"]
            if group not in groups:
                groups[group] = hub.config.ARGS["parser"].add_argument_group(group)
            groups[group].add_argument(*args, **kwargs)
            continue
        if "ex_group" in comps:
            group = comps["ex_group"]
            if group not in ex_groups:
                ex_groups[group] = hub.config.ARGS[
                    "parser"
                ].add_mutually_exclusive_group()
            ex_groups[group].add_argument(*args, **kwargs)
            continue
        if "subcommands" in comps:
            subs = comps["subcommands"]
            if not isinstance(subs, list):
                subs = [subs]
            for sub in subs:
                sparse = hub.config.ARGS["subs"].get(sub)
                if not sparse:
                    # Maybe raise exception here? Malformed config?
                    continue
                sparse.add_argument(*args, **kwargs)
            continue
        hub.config.ARGS["parser"].add_argument(*args, **kwargs)
    return defaults


def parse(
    hub,
    args: List[str] = None,
    namespace: argparse.Namespace = None,
    only_parse_known_arguments: bool = False,
) -> Dict[str, Any]:
    """
    Parse the command line options
    """
    if only_parse_known_arguments:
        opts, unknown_args = hub.config.ARGS["parser"].parse_known_args(args, namespace)
        opts_dict = opts.__dict__
        opts_dict["_unknown_args_"] = unknown_args
    else:
        opts = hub.config.ARGS["parser"].parse_args(args, namespace)
        opts_dict = opts.__dict__
    return opts_dict


def render(hub, cli_args: Dict[str, Any], raw_cli: Dict[str, Any]) -> Dict[str, Any]:
    """
    For options specified as such, take the string passed into the cli and
    render it using the specified render flag
    """
    for key, val in raw_cli.items():
        if key not in cli_args:
            continue
        if "render" not in val:
            continue
        if val["default"] != cli_args[key]:
            # The value was changed, render it
            cli_args[key] = hub.config.render.init.process(val["render"], cli_args[key])
    return cli_args



"""
This sub is used to set up logging for pop projects and injects logging
options into conf making it easy to add robust logging
"""
# Import python libs
import logging


def __init__(hub):
    """
    Set up variables used by the log subsystem
    """
    hub.config.log.LEVELS = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }


def conf(hub, name):
    """
    Return the conf dict for logging, this should be merged OVER by the loaded
    config dict(s)
    """
    # TODO: Make this more robust to handle more logging interfaces
    ldict = {
        "log_file": {
            "default": f"{name}.log",
            "help": "The location of the log file",
            "group": "Logging Options",
        },
        "log_level": {
            "default": "info",
            "help": "Set the log level, either quiet, info, warning, or error",
            "group": "Logging Options",
        },
        "log_fmt_logfile": {
            "default": "%(asctime)s,%(msecs)03d [%(name)-17s][%(levelname)-8s] %(message)s",
            "help": "The format to be given to log file messages",
            "group": "Logging Options",
        },
        "log_fmt_console": {
            "default": "[%(levelname)-8s] %(message)s",
            "help": "The log formatting used in the console",
            "group": "Logging Options",
        },
        "log_datefmt": {
            "default": "%H:%M:%S",
            "help": "The date format to display in the logs",
            "group": "Logging Options",
        },
        "log_plugin": {
            "default": "basic",
            "help": "The logging plugin to use",
            "group": "Logging Options",
        },
    }
    return ldict



"""
Find the conf.py files specificed in sources
"""
# Import python libs
import importlib
import copy
import secrets
import os


def _load_pyimp(hub, imp):
    """
    Load up a python path, parse it and return the conf dataset
    """
    ret = {imp: {}}
    cmod = importlib.import_module(f"{imp}.conf")
    path = os.path.dirname(cmod.__file__)
    for sec in hub.config.SECTIONS:
        ret[imp][sec] = copy.deepcopy(getattr(cmod, sec, {}))
    return path, ret


def load(hub, sources, dyne_names, cli):
    """
    Look over the sources list and find the correct conf.py files
    """
    # Dynamic names
    # First gather the defined sources, they are the autoritative ones
    # Then ditect what the dynamic names are in the source
    # Merged the sources dyne names with any passed dyne names
    # Load up and extend the raw with all of the dynamic names
    raw = {}
    dyne = hub.pop.dyne.get()
    if not isinstance(sources, list):
        sources = [sources]
    for source in sources:
        try:
            path, data = _load_pyimp(hub, source)
        except ImportError:
            continue
        hub.pop.dicts.update(raw, data)
    dnames = set(dyne_names)
    for source in raw:
        for dname in raw[source]["DYNE"]:
            dnames.add(dname)
    for name in dnames:
        if name in dyne:
            if name not in raw:
                raw[name] = {"CONFIG": {}, "CLI_CONFIG": {}}
            if "CLI_CONFIG" in dyne[name]:
                draw = {}
                cli_draw = {}
                for key, val in dyne[name]["CLI_CONFIG"].items():
                    if val.get("dyne"):
                        val["source"] = name
                        cli_draw[key] = val
                hub.pop.dicts.update(raw[cli]["CLI_CONFIG"], cli_draw)
    raw = hub.config.inject.add(cli, raw)
    return raw


def verify(hub, opts):
    """
    Verify that the environment and all named directories in the
    configuration exist
    """
    for imp in opts:
        for key in opts[imp]:
            if key == "root_dir":
                continue
            if key == "config_dir":
                continue
            if key.endswith("_dir"):
                if not os.path.isdir(opts[imp][key]):
                    os.makedirs(opts[imp][key])



"""
Used to take care of the options that end in `_dir`. The assumption is that
`_dir` options need to be treated differently. They need to verified to exist
and they need to be rooted based on the user, root option etc.
"""
# Import python libs
import os


def apply(hub, ret, raw, cli):
    """
    Take the initial defaults and apply the roots system to set the values
    to be either root or in user dotfiles
    """
    default_root = raw.get(cli, {}).get("CONFIG", {}).get("root_dir")
    root_dir = ret.get(cli, {}).get("root_dir")
    os_root = os.path.abspath(os.sep)
    root = os_root
    change = False
    non_priv = False
    if hasattr(os, "geteuid"):
        if not os.geteuid() == 0:
            change = True
            non_priv = True
    if root_dir and root_dir != default_root:
        root = root_dir
        change = True
    if not root.endswith(os.sep):
        root = f"{root}{os.sep}"
    if change:
        for imp in ret:
            for key, val in ret[imp].items():
                if key == "root_dir":
                    continue
                if key.endswith("_dir"):
                    if non_priv:
                        root = os.path.join(os.environ["HOME"], f".{imp}{os.sep}")
                    if imp in val:
                        typename = os.path.basename(
                            os.path.dirname(val[: val.index(imp)])
                        )
                        a_len = len(imp) + 1
                        tgt = os.path.join(
                            root, typename, val[val.index(imp) + a_len :]
                        )
                        ret[imp][key] = tgt
                    else:
                        ret[imp][key] = ret[imp][key].replace(os_root, root, 1)
    return ret



def load(hub, sources, cli=None, dyne_names=None, loader="yaml", parse_cli=True):
    """
    Load up the configs from the integrate system
    """
    if dyne_names is None:
        dyne_names = []
    raw = hub.config.dirs.load(sources, dyne_names, cli)
    import pprint

    print("raw")
    pprint.pprint(raw)
    os_vars = hub.config.os.init.gather(raw)
    cli_args, raw_cli = hub.config.args.gather(raw, cli, parse_cli)
    if cli_args["version"]:
        hub.config.version.run(cli)
    configs = hub.config.file.init.parse(raw, cli, os_vars, cli_args, loader)
    opt = hub.config.order.apply(raw, raw_cli, cli, cli_args, os_vars, configs)
    hub.OPT = hub.pop.data.imap(opt)



"""
Support embedding version number lookup into cli
"""
# Import python libs
import importlib
import sys


CONFIG = {
    "version": {
        "default": False,
        "action": "store_true",
        "help": "Display version information",
    }
}


def run(hub, primary):
    """
    Check the version number and then exit
    """
    mod = importlib.import_module(f"{primary}.version")
    print(f"{primary} {mod.version}")
    sys.exit(0)



# -*- coding: utf-8 -*-
"""
Configuration file core loading functions
"""

# Import python libs
import os
import glob
import fnmatch


def parse(hub, raw, cli, os_vars, cli_args, loader):
    """
    Determine if a config file or a config dir has been set up and load it up!
    """
    # Figure out what config file value to use in this priority
    default = raw[cli]["CONFIG"].get("config", {}).get("default")
    default_dir = raw[cli]["CONFIG"].get("config_dir", {}).get("default")
    os_conf = os_vars.get("config", default)
    os_dir = os_vars.get("config_dir", default_dir)
    conf = cli_args.get("config", os_conf)
    dir_ = cli_args.get("config_dir", os_dir)
    file_opts = {}
    dir_opts = {}
    if dir_:
        file_opts = hub.config.file.init.load_dir(dir_, loader)
    if conf:
        dir_opts = hub.config.file.init.load(conf, loader)
    ret = hub.pop.dicts.update(dir_opts, file_opts)
    return ret


def load(hub, paths, loader, includes=True):
    """
    Load a single configuration file
    """
    opts = {}
    if not isinstance(paths, list):
        paths = [paths]
    add = []
    for fn in paths:
        add.extend(glob.glob(fn))
    paths.extend(add)
    for fn in paths:
        fn_data = getattr(hub, f"config.render.{loader}.load")(fn)
        if includes:
            fn_data = hub.config.file.init.proc_include(fn, fn_data, loader)
        hub.pop.dicts.update(opts, fn_data)
    return opts


def load_dir(
    hub, confdir, loader, includes=True, recurse=True,
):
    """
    Load takes a directory location to scan for configuration files. These
    files will be read in.
    """
    opts = {}
    if not isinstance(confdir, list):
        confdir = [confdir]
    confdirs = []
    for dirs in confdir:
        if not isinstance(dirs, (list, tuple)):
            dirs = [dirs]
        for dir_ in dirs:
            confdirs.extend(glob.glob(dir_))
    paths = []
    for dir_ in confdirs:
        dirpaths = []
        if os.path.isdir(dir_):
            if not recurse:
                for fn_ in os.listdir(dir_):
                    path = os.path.join(dir_, fn_)
                    if os.path.isdir(path):
                        # Don't process directories
                        continue
                    dirpaths.append(path)
            else:
                for root, dirs, files in os.walk(dir_):
                    for fn_ in files:
                        path = os.path.join(root, fn_)
                        dirpaths.append(path)

        # Sort confdir directory paths like:
        # /b.txt
        # /c.txt
        # /a/x.txt
        # /b/x.txt
        paths.extend(sorted(dirpaths, key=lambda p: (p.count(os.path.sep), p)))
    opts = hub.pop.dicts.update(
        opts, hub.config.file.init.load(paths, loader, includes)
    )
    return opts


def proc_include(hub, fn, opts, loader):
    """
    Process include and include_dir
    """
    dirname = os.path.dirname(fn)
    if opts.get("include_dir"):
        idir = opts.pop("include_dir")
        if not idir.startswith(os.path.abspath(os.sep)):
            idir = os.path.join(dirname, idir)
        opts = hub.pop.dicts.update(opts, hub.config.file.init.load_dir(idir, loader))
        hub.config.file.init.proc_include(os.path.join(idir, "f"), opts, loader)
    if opts.get("include"):
        ifn = opts.pop("include")
        if not ifn.startswith(os.path.abspath(os.sep)):
            ifn = os.path.join(dirname, ifn)
        opts = hub.pop.dicts.update(opts, hub.config.file.init.load(ifn, loader))
        hub.config.file.init.proc_include(ifn, opts, loader)
    return opts



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Import python libs
import os
import shutil
from setuptools import setup, Command

NAME = "pop_config"
PYNAME = "pop-config"
DESC = ""

# Version info -- read without importing
_locals = {}
with open("{}/version.py".format(NAME)) as fp:
    exec(fp.read(), None, _locals)
VERSION = _locals["version"]
SETUP_DIRNAME = os.path.dirname(__file__)
if not SETUP_DIRNAME:
    SETUP_DIRNAME = os.getcwd()

with open("README.rst", encoding="utf-8") as f:
    LONG_DESC = f.read()

with open("requirements.txt") as f:
    REQUIREMENTS = f.read().splitlines()


class Clean(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for subdir in (NAME, "tests"):
            for root, dirs, files in os.walk(
                os.path.join(os.path.dirname(__file__), subdir)
            ):
                for dir_ in dirs:
                    if dir_ == "__pycache__":
                        shutil.rmtree(os.path.join(root, dir_))


def discover_packages():
    modules = []
    for package in (NAME,):
        for root, _, files in os.walk(os.path.join(SETUP_DIRNAME, package)):
            pdir = os.path.relpath(root, SETUP_DIRNAME)
            modname = pdir.replace(os.sep, ".")
            modules.append(modname)
    return modules


setup(
    name=PYNAME,
    author="Thomas S Hatch",
    author_email="thatch45@gmail.com",
    url="https://www.saltstack.com",
    version=VERSION,
    install_requires=REQUIREMENTS,
    description=DESC,
    long_description=LONG_DESC,
    long_description_content_type="text/x-rst",
    python_requires=">=3.6",
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 5 - Production/Stable",
    ],
    packages=discover_packages(),
    entry_points={"console_scripts": ["conf = conf.scripts:start",],},
    cmdclass={"clean": Clean},
)



# Import python libs
import logging


def setup(hub, conf):
    """
    Given the configuration data set up the logger
    """
    level = hub.config.log.LEVELS.get(conf["log_level"], logging.INFO)
    root = logging.getLogger("")
    root.setLevel(level)
    cf = logging.Formatter(fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"])
    ch = logging.StreamHandler()
    ch.setLevel(level)
    ch.setFormatter(cf)
    root.addHandler(ch)
    ff = logging.Formatter(fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"])
    fh = logging.FileHandler(conf["log_file"])
    fh.setLevel(level)
    fh.setFormatter(ff)
    root.addHandler(fh)



"""
Read in keys from *NIX like oses - AKA Environement variables
"""
# Import python libs
import os

__virtualname__ = "system"


def __virtual__(hub):
    """
    Don't load on Windows, this is for *nix style platforms
    """
    # TODO: detect if not windows
    return True


def collect(hub, key):
    """
    Collect the option from environment variable if present
    """
    ret = {}
    key = key.upper()
    if key in os.environ:
        return os.environ[key]
    return None
